from dataclasses import dataclass, field
from typing import List

import numpy
from marshmallow import Schema, fields, post_load


class Distribution:
    """
    Base type for all distribution classes.
    """

    def dumpsPacked(self):
        raise Exception(f"'dumpsPacked' has not been implemented in {str(type(self))}")


class NormalSchema(Schema):
    _type = fields.String(default="IncrementalInference.PackedNormal")
    mu = fields.Float()
    sigma = fields.Float()

    @post_load
    def marshal(self, data, **kwargs):
        return Normal(**data)

    class Meta:
        ordered = True


@dataclass
class Normal(Distribution):
    mu: float
    sigma: float
    _type: str = "IncrementalInference.PackedNormal"

    def dumps(self):
        return NormalSchema().dumps(self)

    def dump(self):
        return NormalSchema().dump(self)

    @staticmethod
    def load(data):
        return NormalSchema().load(data)


class RayleighSchema(Schema):
    _type = fields.String(default="IncrementalInference.PackedRayleigh")
    sigma = fields.Float()

    @post_load
    def marshal(self, data, **kwargs):
        return Normal(**data)

    class Meta:
        ordered = True


@dataclass
class Rayleigh(Distribution):
    sigma: float
    _type: str = "IncrementalInference.PackedRayleigh"

    def dumps(self):
        return RayleighSchema().dumps(self)

    def dump(self):
        return RayleighSchema().dump(self)

    @staticmethod
    def load(data):
        return RayleighSchema().load(data)


class FullNormalSchema(Schema):
    _type = fields.String(default="IncrementalInference.PackedFullNormal")
    mu = fields.List(fields.Float)
    cov = fields.Method("get_cov", "set_cov", required=True)

    def get_cov(self, obj):
        return obj.cov.flatten().tolist()

    def set_cov(self, obj):
        raise Exception("Deserialization not supported yet.")  # obj.cov.reshape(3,3)?

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return FullNormal(**data)


@dataclass
class FullNormal(Distribution):
    # TODO: Remove the default initializers.
    mu: numpy.ndarray = field(default_factory=lambda: numpy.zeros(3))
    cov: numpy.ndarray = field(default_factory=lambda: numpy.diag([0.1, 0.1, 0.1]))
    _type: str = "IncrementalInference.PackedFullNormal"

    def dumps(self):
        return FullNormalSchema().dumps(self)

    def dump(self):
        return FullNormalSchema().dump(self)

    @staticmethod
    def load(data):
        return FullNormalSchema().load(data)


@dataclass
class Uniform(Distribution):
    a: float
    b: float

    def dumps(self):
        return UniformSchema().dumps(self)

    def dump(self):
        return UniformSchema().dump(self)

    @staticmethod
    def load(data):
        return UniformSchema().load(data)


class UniformSchema(Schema):
    _type = fields.String(default="IncrementalInference.PackedUniform")
    a = fields.Float()
    b = fields.Float()
    PackedSamplableTypeJSON = fields.String(
        default="IncrementalInference.PackedUniform"
    )

    class Meta:
        ordered = True


@dataclass
class Categorical(Distribution):
    p: List[float]

    def dumps(self):
        return CategoricalSchema().dumps(self)

    def dump(self):
        return CategoricalSchema().dump(self)

    @staticmethod
    def load(data):
        return CategoricalSchema().load(data)


class CategoricalSchema(Schema):
    _type = fields.String(default="IncrementalInference.PackedCategorical")
    p = fields.List(fields.Float)

    class Meta:
        ordered = True



from collections import OrderedDict
from dataclasses import dataclass
from typing import List

import numpy as np
from marshmallow import Schema, fields, post_load

from navability.entities.factor.distributions import Categorical, Distribution


@dataclass
class InferenceType:
    """
    Base type for all factor classes.
    """

    pass


@dataclass
class Prior(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


class ZSchema(Schema):
    Z = fields.Method("get_Z", "set_Z", required=True)

    class Meta:
        ordered = True

    def get_Z(self, obj):
        return obj.Z.dump()

    def set_Z(self, obj):
        raise Exception("This has not been implemented yet.")

    # @post_load
    # def load(self, data, **kwargs):
    #     return PriorPose2(**data)


"""
Create a ContinousScalar->ContinousScalar (also known as Pose1->Pose1)
factor with a distribution Z representing the 1D relationship
between the variables, e.g. `Normal(1.0, 0.1)`.
"""


@dataclass
class LinearRelative(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a prior factor for a Pose2 with a distribution Z representing
(x,y,theta) prior information, e.g. `FullNormal([0.0, 0.0, 0.0],
diagm([0.01, 0.01, 0.01]))`.
"""


@dataclass
class PriorPose2(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a prior factor for a Point2 with a distribution Z representing (x,y) prior
information, e.g. `FullNormal([0.0, 0.0.0], diag([0.01, 0.01]))`.
"""


@dataclass
class PriorPoint2(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a Pose2->Pose2 factor with a distribution Z representing the
(x,y,theta) relationship between the variables, e.g.
`FullNormal([1,0,0.3333*π], diag([0.01,0.01,0.01]))`.
"""


@dataclass
class Pose2Pose2(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a Point2->Point2 range factor with a 1D distribution:
- range: The range from the pose to the point.
"""


@dataclass
class Point2Point2Range(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a Pose2->Point2 range factor with a 1D distribution:
- range: The range from the pose to the point.
"""


@dataclass
class Pose2Point2Range(InferenceType):
    Z: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(Z={str(self.Z)})>"

    def dump(self):
        return ZSchema().dump(self)

    def dumps(self):
        return ZSchema().dumps(self)

    # TODO: Deserializing this.


"""
Create a Pose2->Point2 bearing+range factor with 1D distributions:
- bearing: The bearing from the pose to the point.
- range: The range from the pose to the point.
"""


@dataclass
class Pose2Point2BearingRange(InferenceType):
    bearing: Distribution
    range: Distribution

    def __repr__(self):
        return f"<{self.__class__.__name__}(bearing={str(self.bearing)}, range={str(self.range)})>"  # noqa: E501, B950

    def dump(self):
        return Pose2Point2BearingRangeSchema().dump(self)

    def dumps(self):
        return Pose2Point2BearingRangeSchema().dumps(self)

    # TODO: Deserializing this.


class Pose2Point2BearingRangeSchema(Schema):
    bearstr = fields.Method("get_bearstr", required=True)
    rangstr = fields.Method("get_rangstr", required=True)

    class Meta:
        ordered = True

    def get_bearstr(self, obj):
        return obj.bearing.dump()

    def get_rangstr(self, obj):
        return obj.range.dump()

    # def set_Z(self, obj):
    #     raise Exception("This has not been implemented yet.")


"""
Create a AprilTags factor that directly relates a Pose2 to the
information from an AprilTag reading. Corners need to be provided,
homography and tag length are defaulted and can be overwritten.
"""


@dataclass
class Pose2AprilTag4Corners(InferenceType):
    corners: np.ndarray
    homography: np.ndarray
    K: np.ndarray
    taglength: float
    id: int
    _type: str = "/application/JuliaLang/PackedPose2AprilTag4Corners"

    def dump(self):
        return Pose2AprilTag4CornersSchema().dump(self)

    def dumps(self):
        return Pose2AprilTag4CornersSchema().dumps(self)

    # TODO: Deserializing this.


class Pose2AprilTag4CornersSchema(Schema):
    corners = fields.List(fields.Float, required=True)
    homography = fields.List(fields.Float, required=True)
    K = fields.List(fields.Float, required=True)
    taglength = fields.Float(required=True)
    id = fields.Int(required=True)
    _type = fields.String(default="/application/JuliaLang/PackedPose2AprilTag4Corners")

    class Meta:
        ordered = True

    @post_load
    def marshal(self, data, **kwargs):
        return Pose2AprilTag4Corners(**data)


@dataclass
class Mixture(InferenceType):
    mechanics: type
    components: "OrderedDict[str, Distribution]"
    diversity: Categorical
    dims: int  # Internal factor dimensions

    def __init__(
        self,
        mechanics: type,
        components: "OrderedDict[str, Distribution]",
        probabilities: List[float],
        dims: int,
    ):
        """Create a Mixture factor type with an underlying factor type, a named set of
         distributions that should be mixed, the probabilities of each distribution
        (the mix), and the dimensions of the underlying factor
        (e.g. OrderedDict([("hypo1", Normal(0, 2)), ("hypo2", Uniform(30, 55))]))
        ContinuousScalar=1, Pose2Pose2=3, etc.).

        Args:
            mechanics (type): The underlying factor type, e.g. Pose2Pose2.
            components (OrderedDict[Distribution]): The named set of distributions that should
            be mixed, e.g. OrderedDict([("hypo1", Normal(0, 2)), ("hypo2", Uniform(30, 55))])
            probabilities (List[float]): The probabilities of each
            distribution (the mix), e.g. [0.4, 0.6]
            dims (int): The dimensions of the underlying factor, e.g. for Pose2Pose2 it's 3
        """  # noqa: E501, B950
        self.mechanics = mechanics
        self.components = components
        self.diversity = Categorical(probabilities)
        self.dims = dims

    def dump(self):
        return MixtureSchema().dump(self)

    def dumps(self):
        return MixtureSchema().dumps(self)

    # TODO: Deserializing this.


class MixtureSchema(Schema):
    N = fields.Method("get_N")
    # TODO: Need to deserialize these at some point in the future.
    F_ = fields.Method("get_F_")
    S = fields.Method("get_S")
    components = fields.Method("get_components")
    diversity = fields.Method("get_diversity")

    class Meta:
        ordered = True

    def get_N(self, obj):
        return len(obj.components)

    def get_F_(self, obj):
        # Not certain this is correct, we will need to validate as we use it.
        return f"Packed{obj.mechanics.__name__}"

    def get_S(self, obj):
        return list(obj.components.keys())

    def get_components(self, obj):
        return list(c.dump() for c in obj.components.values())

    def get_diversity(self, obj):
        return obj.diversity.dump()

    @post_load
    def marshal(self, data, **kwargs):
        return Mixture(**data)



from typing import List

from gql import gql

from navability.common.queries import GQL_GETSTATUSLATEST, GQL_GETSTATUSMESSAGES
from navability.entities.navabilityclient import NavAbilityClient, QueryOptions
from navability.entities.statusmessage import StatusMessageSchema


async def getStatusMessages(navAbilityClient: NavAbilityClient, id: str):
    """Get all the statuses for a request.

    Args:
        navAbilityClient (NavAbilityClient): The NavAbility client.
        id (str): The ID of the request that you want the statuses on.
    """
    statusMessages = await navAbilityClient.query(
        QueryOptions(gql(GQL_GETSTATUSMESSAGES), {"id": id})
    )
    schema = StatusMessageSchema(many=True)
    return schema.load(statusMessages["statusMessages"])


async def getStatusLatest(navAbilityClient: NavAbilityClient, id: str):
    """Get the latest status message for a request.

    Args:
        navAbilityClient (NavAbilityClient): The NavAbility client.
        id (str): The ID of the request that you want the latest status on.
    """
    statusMessages = await navAbilityClient.query(
        QueryOptions(gql(GQL_GETSTATUSLATEST), {"id": id})
    )
    schema = StatusMessageSchema()
    return schema.load(statusMessages["statusLatest"])


async def getStatusesLatest(navAbilityClient: NavAbilityClient, ids: List[str]):
    """Helper function to get a dictionary of all latest statues for a list of results.

    Args:
        navAbilityClient (NavAbilityClient): The NavAbility client.
        ids (List[str]): A list of the IDS that you want statuses on.
    """
    return {r: await getStatusLatest(navAbilityClient, r) for r in ids}



import time
from typing import List

from navability.entities import NavAbilityClient
from navability.services import getStatusesLatest


async def waitForCompletion(
    navAbilityClient: NavAbilityClient,
    requestIds: List[str],
    maxSeconds: int = 60,
    expectedStatuses: List[str] = None,
    exceptionMessage: str = "Requests did not complete in time",
):
    """Wait for the requests to complete, poll until done.

    Args:
        requestIds (List[str]): The request IDs that should be polled.
        maxSeconds (int, optional): Maximum wait time. Defaults to 60.
        expectedStatus (str, optional): Expected status message per request.
            Defaults to "Complete".
    """
    if expectedStatuses is None:
        expectedStatuses = ["Complete", "Failed"]
    wait_time = maxSeconds
    tasksComplete = False
    while not tasksComplete:
        statuses = (await getStatusesLatest(navAbilityClient, requestIds)).values()
        tasksComplete = all(s.state in expectedStatuses for s in statuses)
        if tasksComplete:
            break
        else:
            time.sleep(2)
            wait_time -= 2
            if wait_time <= 0:
                raise Exception(exceptionMessage)
